@@ -1,27 +1,8 @@
-extern crate base64;
-extern crate crossbeam_utils;
-extern crate env_logger;
-extern crate flate2;
-extern crate jsonwebtoken as jwt;
-
-#[cfg(not(target_os = "freebsd"))]
-extern crate libmount;
-
 #[macro_use]
 extern crate log;
-extern crate nix;
-extern crate openssl;
-extern crate rand;
-extern crate reqwest;
-extern crate sccache;
-#[macro_use]
-extern crate serde_derive;
-extern crate serde_json;
-extern crate syslog;
-extern crate tar;
-extern crate void;
 
 use anyhow::{bail, Context, Error, Result};
+use base64::Engine;
 use rand::{rngs::OsRng, RngCore};
 use sccache::config::{
     scheduler as scheduler_config, server as server_config, INSECURE_DIST_CLIENT_TOKEN,
@@ -35,6 +16,7 @@ use sccache::dist::{
 };
 use sccache::util::daemonize;
 use sccache::util::BASE64_URL_SAFE_ENGINE;
+use serde::{Deserialize, Serialize};
 use std::collections::{btree_map, BTreeMap, HashMap, HashSet};
 use std::env;
 use std::io;
@@ -61,6 +43,17 @@ pub const INSECURE_DIST_SERVER_TOKEN: &str = "dangerously_insecure_server";
 fn main() {
     init_logging();
 
+    let incr_env_strs = ["CARGO_BUILD_INCREMENTAL", "CARGO_INCREMENTAL"];
+    incr_env_strs
+        .iter()
+        .for_each(|incr_str| match env::var(incr_str) {
+            Ok(incr_val) if incr_val == "1" => {
+                println!("sccache: increment compilation is  prohibited.");
+                std::process::exit(1);
+            }
+            _ => (),
+        });
+
     let command = match cmdline::try_parse_from(env::args()) {
         Ok(cmd) => cmd,
         Err(e) => match e.downcast::<clap::error::Error>() {
@@ -144,7 +137,7 @@ fn run(command: Command) -> Result<i32> {
             let mut bytes = vec![0; num_bytes];
             OsRng.fill_bytes(&mut bytes);
             // As long as it can be copied, it doesn't matter if this is base64 or hex etc
-            println!("{}", base64::encode_engine(&bytes, &BASE64_URL_SAFE_ENGINE));
+            println!("{}", BASE64_URL_SAFE_ENGINE.encode(bytes));
             Ok(0)
         }
         Command::Auth(AuthSubcommand::JwtHS256ServerToken {
@@ -152,7 +145,7 @@ fn run(command: Command) -> Result<i32> {
             server_id,
         }) => {
             let header = jwt::Header::new(jwt::Algorithm::HS256);
-            let secret_key = base64::decode_engine(&secret_key, &BASE64_URL_SAFE_ENGINE)?;
+            let secret_key = BASE64_URL_SAFE_ENGINE.decode(&secret_key)?;
             let token = create_jwt_server_token(server_id, &header, &secret_key)
                 .context("Failed to create server token")?;
             println!("{}", token);
@@ -197,7 +190,8 @@ fn run(command: Command) -> Result<i32> {
                     Box::new(move |server_token| check_server_token(server_token, &token))
                 }
                 scheduler_config::ServerAuth::JwtHS256 { secret_key } => {
-                    let secret_key = base64::decode_engine(&secret_key, &BASE64_URL_SAFE_ENGINE)
+                    let secret_key = BASE64_URL_SAFE_ENGINE
+                        .decode(&secret_key)
                         .context("Secret key base64 invalid")?;
                     if secret_key.len() != 256 / 8 {
                         bail!("Size of secret key incorrect")
@@ -223,7 +217,8 @@ fn run(command: Command) -> Result<i32> {
                 check_client_auth,
                 check_server_auth,
             );
-            void::unreachable(http_scheduler.start()?);
+            http_scheduler.start()?;
+            unreachable!();
         }
 
         Command::Server(server_config::Config {
@@ -299,7 +294,8 @@ fn run(command: Command) -> Result<i32> {
                 server,
             )
             .context("Failed to create sccache HTTP server instance")?;
-            void::unreachable(http_server.start()?)
+            http_server.start()?;
+            unreachable!();
         }
     }
 }
