@@ -15,7 +15,8 @@
 
 use opendal::Operator;
 
-use opendal::services::azblob;
+use opendal::layers::LoggingLayer;
+use opendal::services::Azblob;
 
 use crate::errors::*;
 
@@ -23,10 +24,13 @@ pub struct AzureBlobCache;
 
 impl AzureBlobCache {
     pub fn build(connection_string: &str, container: &str, key_prefix: &str) -> Result<Operator> {
-        let mut builder = azblob::Builder::from_connection_string(connection_string)?;
+        let mut builder = Azblob::from_connection_string(connection_string)?;
         builder.container(container);
         builder.root(key_prefix);
 
-        Ok(builder.build()?.into())
+        let op = Operator::new(builder)?
+            .layer(LoggingLayer::default())
+            .finish();
+        Ok(op)
     }
 }
