@@ -18,6 +18,7 @@ use opendal::layers::LoggingLayer;
 use opendal::services::Redis;
 use opendal::Operator;
 use std::collections::HashMap;
+use std::time::Duration;
 use url::Url;
 
 /// A cache that stores entries in a Redis.
@@ -25,13 +26,16 @@ pub struct RedisCache;
 
 impl RedisCache {
     /// Create a new `RedisCache`.
-    pub fn build(url: &str) -> Result<Operator> {
+    pub fn build(url: &str, ttl: u64) -> Result<Operator> {
         let parsed = Url::parse(url)?;
 
         let mut builder = Redis::default();
         builder.endpoint(parsed.as_str());
         builder.username(parsed.username());
         builder.password(parsed.password().unwrap_or_default());
+        if ttl != 0 {
+            builder.default_ttl(Duration::from_secs(ttl));
+        }
 
         let options: HashMap<_, _> = parsed
             .query_pairs()
