@@ -58,7 +58,7 @@ impl Default for StatsFormat {
 /// A specific command to run.
 pub enum Command {
     /// Show cache statistics and exit.
-    ShowStats(StatsFormat),
+    ShowStats(StatsFormat, bool),
     /// Run background server.
     InternalStartServer,
     /// Start background server as a subprocess.
@@ -84,6 +84,7 @@ pub enum Command {
         /// The environment variables to use for execution.
         env_vars: Vec<(OsString, OsString)>,
     },
+    DebugPreprocessorCacheEntries,
 }
 
 fn flag_infer_long_and_short(name: &'static str) -> Arg {
@@ -124,9 +125,15 @@ fn get_clap_command() -> clap::Command {
             flag_infer_long_and_short("show-stats")
                 .help("show cache statistics")
                 .action(ArgAction::SetTrue),
+            flag_infer_long("show-adv-stats")
+                .help("show advanced cache statistics")
+                .action(ArgAction::SetTrue),
             flag_infer_long("start-server")
                 .help("start background server")
                 .action(ArgAction::SetTrue),
+            flag_infer_long("debug-preprocessor-cache")
+                .help("show all preprocessor cache entries")
+                .action(ArgAction::SetTrue),
             flag_infer_long("stop-server")
                 .help("stop background server")
                 .action(ArgAction::SetTrue),
@@ -143,7 +150,7 @@ fn get_clap_command() -> clap::Command {
                 .help("package toolchain for distributed compilation")
                 .value_parser(clap::value_parser!(PathBuf))
                 .num_args(2)
-                .value_names(&["EXE", "OUT"]),
+                .value_names(["EXE", "OUT"]),
             flag_infer_long("stats-format")
                 .help("set output format of statistics")
                 .value_name("FMT")
@@ -156,10 +163,12 @@ fn get_clap_command() -> clap::Command {
         ])
         .group(
             ArgGroup::new("one_and_only_one")
-                .args(&[
+                .args([
                     "dist-auth",
+                    "debug-preprocessor-cache",
                     "dist-status",
                     "show-stats",
+                    "show-adv-stats",
                     "start-server",
                     "stop-server",
                     "zero-stats",
@@ -246,9 +255,17 @@ pub fn try_parse() -> Result<Command> {
                     .get_one("stats-format")
                     .cloned()
                     .expect("There is a default value");
-                Ok(Command::ShowStats(fmt))
+                Ok(Command::ShowStats(fmt, false))
+            } else if matches.get_flag("show-adv-stats") {
+                let fmt = matches
+                    .get_one("stats-format")
+                    .cloned()
+                    .expect("There is a default value");
+                Ok(Command::ShowStats(fmt, true))
             } else if matches.get_flag("start-server") {
                 Ok(Command::StartServer)
+            } else if matches.get_flag("debug-preprocessor-cache") {
+                Ok(Command::DebugPreprocessorCacheEntries)
             } else if matches.get_flag("stop-server") {
                 Ok(Command::StopServer)
             } else if matches.get_flag("zero-stats") {
@@ -275,11 +292,19 @@ pub fn try_parse() -> Result<Command> {
                 // If we're running under rr, avoid the `LD_PRELOAD` bits, as it will
                 // almost surely do the wrong thing, as the compiler gets executed
                 // in a different process tree.
-                //
-                // FIXME: Maybe we should strip out `LD_PRELOAD` always?
-                if env::var_os("RUNNING_UNDER_RR").is_some() {
-                    env_vars.retain(|(k, _v)| k != "LD_PRELOAD" && k != "RUNNING_UNDER_RR");
-                }
+                env_vars.retain(|(k, _v)| {
+                    k != "LD_PRELOAD"
+                        && k != "RUNNING_UNDER_RR"
+                        && k != "HOSTNAME"
+                        && k != "PWD"
+                        && k != "HOST"
+                        && k != "RPM_BUILD_ROOT"
+                        && k != "SOURCE_DATE_EPOCH"
+                        && k != "RPM_PACKAGE_RELEASE"
+                        && k != "MINICOM"
+                        && k != "DESTDIR"
+                        && k != "RPM_PACKAGE_VERSION"
+                });
 
                 let cmd = matches
                     .get_many("CMD")
