@@ -12,14 +12,15 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+use crate::cache::storage_from_config;
 use crate::client::{connect_to_server, connect_with_retry, ServerConnection};
 use crate::cmdline::{Command, StatsFormat};
 use crate::compiler::ColorMode;
-use crate::config::Config;
+use crate::config::{default_disk_cache_dir, Config};
 use crate::jobserver::Client;
 use crate::mock_command::{CommandChild, CommandCreatorSync, ProcessCommandCreator, RunCommand};
 use crate::protocol::{Compile, CompileFinished, CompileResponse, Request, Response};
-use crate::server::{self, DistInfo, ServerInfo, ServerStartup};
+use crate::server::{self, DistInfo, ServerInfo, ServerStartup, ServerStats};
 use crate::util::daemonize;
 use byteorder::{BigEndian, ByteOrder};
 use fs::{File, OpenOptions};
@@ -37,6 +38,7 @@ use std::time::Duration;
 use strip_ansi_escapes::Writer;
 use tokio::io::AsyncReadExt;
 use tokio::runtime::Runtime;
+use walkdir::WalkDir;
 use which::which_in;
 
 use crate::errors::*;
@@ -87,6 +89,15 @@ fn run_server_process(startup_timeout: Option<Duration>) -> Result<ServerStartup
     let runtime = Runtime::new()?;
     let exe_path = env::current_exe()?;
     let workdir = exe_path.parent().expect("executable path has no parent?!");
+
+    // Spawn a blocking task to bind the Unix socket. Note that the socket
+    // must be bound before spawning `_child` below to avoid a race between
+    // the parent binding the socket and the child connecting to it.
+    let listener = {
+        let _guard = runtime.enter();
+        tokio::net::UnixListener::bind(&socket_path)?
+    };
+
     let _child = process::Command::new(&exe_path)
         .current_dir(workdir)
         .env("SCCACHE_START_SERVER", "1")
@@ -95,7 +106,6 @@ fn run_server_process(startup_timeout: Option<Duration>) -> Result<ServerStartup
         .spawn()?;
 
     let startup = async move {
-        let listener = tokio::net::UnixListener::bind(&socket_path)?;
         let (socket, _) = listener.accept().await?;
 
         read_server_startup_status(socket).await
@@ -324,13 +334,13 @@ fn connect_or_start_server(
 }
 
 /// Send a `ZeroStats` request to the server, and return the `ServerInfo` request if successful.
-pub fn request_zero_stats(mut conn: ServerConnection) -> Result<ServerInfo> {
+pub fn request_zero_stats(mut conn: ServerConnection) -> Result<()> {
     debug!("request_stats");
     let response = conn.request(Request::ZeroStats).context(
         "failed to send zero statistics command to server or failed to receive response",
     )?;
-    if let Response::Stats(stats) = response {
-        Ok(*stats)
+    if let Response::ZeroStats = response {
+        Ok(())
     } else {
         bail!("Unexpected server response!")
     }
@@ -339,9 +349,9 @@ pub fn request_zero_stats(mut conn: ServerConnection) -> Result<ServerInfo> {
 /// Send a `GetStats` request to the server, and return the `ServerInfo` request if successful.
 pub fn request_stats(mut conn: ServerConnection) -> Result<ServerInfo> {
     debug!("request_stats");
-    let response = conn
-        .request(Request::GetStats)
-        .context("Failed to send data to or receive data from server")?;
+    let response = conn.request(Request::GetStats).context(
+        "Failed to send data to or receive data from server. Mismatch of client/server versions?",
+    )?;
     if let Response::Stats(stats) = response {
         Ok(*stats)
     } else {
@@ -604,15 +614,41 @@ pub fn run_command(cmd: Command) -> Result<i32> {
     let startup_timeout = config.server_startup_timeout;
 
     match cmd {
-        Command::ShowStats(fmt) => {
+        Command::ShowStats(fmt, advanced) => {
             trace!("Command::ShowStats({:?})", fmt);
-            let srv = connect_or_start_server(get_port(), startup_timeout)?;
-            let stats = request_stats(srv).context("failed to get stats from server")?;
+            let stats = match connect_to_server(get_port()) {
+                Ok(srv) => request_stats(srv).context("failed to get stats from server")?,
+                // If there is no server, spawning a new server would start with zero stats
+                // anyways, so we can just return (mostly) empty stats directly.
+                Err(_) => {
+                    let runtime = Runtime::new()?;
+                    let storage = storage_from_config(config, runtime.handle()).ok();
+                    runtime.block_on(ServerInfo::new(ServerStats::default(), storage.as_deref()))?
+                }
+            };
             match fmt {
-                StatsFormat::Text => stats.print(),
+                StatsFormat::Text => stats.print(advanced),
                 StatsFormat::Json => serde_json::to_writer(&mut io::stdout(), &stats)?,
             }
         }
+        Command::DebugPreprocessorCacheEntries => {
+            trace!("Command::DebugPreprocessorCacheEntries");
+            let entries_dir = default_disk_cache_dir().join("preprocessor");
+            for entry in WalkDir::new(entries_dir).sort_by_file_name().into_iter() {
+                let preprocessor_cache_entry_file = entry?;
+                let path = preprocessor_cache_entry_file.path();
+                if !path.is_file() {
+                    continue;
+                }
+                println!("=========================");
+                println!("Showing preprocessor entry file {}", &path.display());
+                let contents = std::fs::read(path)?;
+                let preprocessor_cache_entry =
+                    crate::compiler::PreprocessorCacheEntry::read(&contents)?;
+                println!("{:#?}", preprocessor_cache_entry);
+                println!("=========================");
+            }
+        }
         Command::InternalStartServer => {
             trace!("Command::InternalStartServer");
             if env::var("SCCACHE_ERROR_LOG").is_ok() {
@@ -647,13 +683,13 @@ pub fn run_command(cmd: Command) -> Result<i32> {
             println!("Stopping sccache server...");
             let server = connect_to_server(get_port()).context("couldn't connect to server")?;
             let stats = request_shutdown(server)?;
-            stats.print();
+            stats.print(false);
         }
         Command::ZeroStats => {
             trace!("Command::ZeroStats");
             let conn = connect_or_start_server(get_port(), startup_timeout)?;
-            let stats = request_zero_stats(conn).context("couldn't zero stats on server")?;
-            stats.print();
+            request_zero_stats(conn).context("couldn't zero stats on server")?;
+            eprintln!("Statistics zeroed.");
         }
         #[cfg(feature = "dist-client")]
         Command::DistAuth => {
