@@ -13,17 +13,18 @@
 // limitations under the License.
 
 use crate::compiler::args::*;
-use crate::compiler::c::{
-    ArtifactDescriptor, CCompilerImpl, CCompilerKind, Language, ParsedArguments,
-};
-use crate::compiler::{clang, Cacheable, ColorMode, CompileCommand, CompilerArguments};
+use crate::compiler::c::{ArtifactDescriptor, CCompilerImpl, CCompilerKind, ParsedArguments};
+use crate::compiler::{clang, Cacheable, ColorMode, CompileCommand, CompilerArguments, Language};
 use crate::mock_command::{CommandCreatorSync, RunCommand};
 use crate::util::{run_input_output, OsStrExt};
 use crate::{counted_array, dist};
+use async_trait::async_trait;
+use fs::File;
+use fs_err as fs;
 use log::Level::Trace;
 use std::collections::HashMap;
+use std::env;
 use std::ffi::OsString;
-use std::fs::File;
 use std::io::Read;
 use std::path::{Path, PathBuf};
 use std::process;
@@ -66,10 +67,16 @@ impl CCompilerImpl for Gcc {
         env_vars: &[(OsString, OsString)],
         may_dist: bool,
         rewrite_includes_only: bool,
+        preprocessor_cache_mode: bool,
     ) -> Result<process::Output>
     where
         T: CommandCreatorSync,
     {
+        let ignorable_whitespace_flags = if preprocessor_cache_mode {
+            vec![]
+        } else {
+            vec!["-P".to_string()]
+        };
         preprocess(
             creator,
             executable,
@@ -79,7 +86,7 @@ impl CCompilerImpl for Gcc {
             may_dist,
             self.kind(),
             rewrite_includes_only,
-            vec!["-P".to_string()],
+            ignorable_whitespace_flags,
         )
         .await
     }
@@ -119,6 +126,8 @@ ArgData! { pub
     PreprocessorArgumentFlag,
     PreprocessorArgument(OsString),
     PreprocessorArgumentPath(PathBuf),
+    // Used for arguments that shouldn't affect the computed hash
+    Unhashed(OsString),
     DoCompilation,
     Output(PathBuf),
     NeedDepTarget,
@@ -173,6 +182,7 @@ counted_array!(pub static ARGS: [ArgInfo<ArgData>; _] = [
     take_arg!("-U", OsString, CanBeSeparated, PassThrough),
     take_arg!("-V", OsString, Separated, PassThrough),
     flag!("-Werror=pedantic", PedanticFlag),
+    take_arg!("-Wp", OsString, Concatenated(','), PreprocessorArgument),
     flag!("-Wpedantic", PedanticFlag),
     take_arg!("-Xassembler", OsString, Separated, PassThrough),
     take_arg!("-Xlinker", OsString, Separated, PassThrough),
@@ -244,10 +254,12 @@ where
 {
     let mut output_arg = None;
     let mut input_arg = None;
+    let mut double_dash_input = false;
     let mut dep_target = None;
     let mut dep_flag = OsString::from("-MT");
     let mut common_args = vec![];
     let mut arch_args = vec![];
+    let mut unhashed_args = vec![];
     let mut preprocessor_args = vec![];
     let mut dependency_args = vec![];
     let mut extra_hash_files = vec![];
@@ -270,12 +282,20 @@ where
     let mut outputs_gcno = false;
     let mut xclangs: Vec<OsString> = vec![];
     let mut color_mode = ColorMode::Auto;
+    let mut seen_arch = None;
+    let dont_cache_multiarch = env::var("SCCACHE_CACHE_MULTIARCH").is_err();
 
     // Custom iterator to expand `@` arguments which stand for reading a file
     // and interpreting it as a list of more arguments.
     let it = ExpandIncludeFile::new(cwd, arguments);
 
-    for arg in ArgsIter::new(it, arg_info) {
+    let mut too_hard_for_preprocessor_cache_mode = false;
+
+    let mut args_iter = ArgsIter::new(it, arg_info);
+    if kind == CCompilerKind::Clang {
+        args_iter = args_iter.with_double_dashes();
+    }
+    for arg in args_iter {
         let arg = try_or_cannot_cache!(arg, "argument parse");
         // Check if the value part of this argument begins with '@'. If so, we either
         // failed to expand it, or it was a concatenated argument - either way, bail.
@@ -348,20 +368,40 @@ where
             | Some(PreprocessorArgument(_))
             | Some(PreprocessorArgumentPath(_))
             | Some(PassThrough(_))
-            | Some(PassThroughPath(_)) => {}
+            | Some(PassThroughPath(_))
+            | Some(Unhashed(_)) => {}
             Some(Language(lang)) => {
                 language = match lang.to_string_lossy().as_ref() {
                     "c" => Some(Language::C),
+                    "c-header" => Some(Language::CHeader),
                     "c++" => Some(Language::Cxx),
+                    "c++-header" => Some(Language::CxxHeader),
                     "objective-c" => Some(Language::ObjectiveC),
                     "objective-c++" => Some(Language::ObjectiveCxx),
                     "cu" => Some(Language::Cuda),
+                    "rs" => Some(Language::Rust),
+                    "cuda" => Some(Language::Cuda),
                     _ => cannot_cache!("-x"),
                 };
             }
-            Some(Arch(_)) => {}
+            Some(Arch(arch)) => {
+                match seen_arch {
+                    Some(s) if &s != arch && dont_cache_multiarch => {
+                        cannot_cache!(
+                            "multiple different -arch, and SCCACHE_CACHE_MULTIARCH not set"
+                        )
+                    }
+                    _ => {}
+                };
+                seen_arch = Some(arch.clone());
+            }
             Some(XClang(s)) => xclangs.push(s.clone()),
             None => match arg {
+                Argument::Raw(ref val) if val == "--" => {
+                    if input_arg.is_none() {
+                        double_dash_input = true;
+                    }
+                }
                 Argument::Raw(ref val) => {
                     if input_arg.is_some() {
                         multiple_input = true;
@@ -387,14 +427,22 @@ where
             | Some(PassThroughFlag)
             | Some(PassThrough(_))
             | Some(PassThroughPath(_)) => &mut common_args,
+            Some(Unhashed(_)) => &mut unhashed_args,
             Some(Arch(_)) => &mut arch_args,
             Some(ExtraHashFile(path)) => {
                 extra_hash_files.push(cwd.join(path));
                 &mut common_args
             }
-            Some(PreprocessorArgumentFlag)
-            | Some(PreprocessorArgument(_))
-            | Some(PreprocessorArgumentPath(_)) => &mut preprocessor_args,
+            Some(PreprocessorArgument(_)) => {
+                too_hard_for_preprocessor_cache_mode = match arg.flag_str() {
+                    Some(s) => s == "-Xpreprocessor" || s == "-Wp",
+                    _ => false,
+                };
+                &mut preprocessor_args
+            }
+            Some(PreprocessorArgumentFlag) | Some(PreprocessorArgumentPath(_)) => {
+                &mut preprocessor_args
+            }
             Some(DepArgumentPath(_)) | Some(NeedDepTarget) => &mut dependency_args,
             Some(DoCompilation) | Some(Language(_)) | Some(Output(_)) | Some(XClang(_))
             | Some(DepTarget(_)) => continue,
@@ -438,7 +486,10 @@ where
                 .unwrap_or("Can't handle complex arguments through clang",)),
             None => match arg {
                 Argument::Raw(_) if follows_plugin_arg => &mut common_args,
-                Argument::Raw(_) => cannot_cache!("Can't handle Raw arguments with -Xclang"),
+                Argument::Raw(flag) => cannot_cache!(
+                    "Can't handle Raw arguments with -Xclang",
+                    flag.to_str().unwrap_or("").to_string()
+                ),
                 Argument::UnknownFlag(flag) => {
                     cannot_cache!(
                         "Can't handle UnknownFlag arguments with -Xclang",
@@ -454,6 +505,7 @@ where
             | Some(PassThrough(_))
             | Some(PassThroughFlag)
             | Some(PassThroughPath(_)) => &mut common_args,
+            Some(Unhashed(_)) => &mut unhashed_args,
             Some(ExtraHashFile(path)) => {
                 extra_hash_files.push(cwd.join(path));
                 &mut common_args
@@ -563,6 +615,7 @@ where
 
     CompilerArguments::Ok(ParsedArguments {
         input: input.into(),
+        double_dash_input,
         language,
         compilation_flag,
         depfile: None,
@@ -571,14 +624,30 @@ where
         preprocessor_args,
         common_args,
         arch_args,
+        unhashed_args,
         extra_hash_files,
         msvc_show_includes: false,
         profile_generate,
         color_mode,
         suppress_rewrite_includes_only,
+        too_hard_for_preprocessor_cache_mode,
     })
 }
 
+fn language_to_gcc_arg(lang: Language) -> Option<&'static str> {
+    match lang {
+        Language::C => Some("c"),
+        Language::CHeader => Some("c-header"),
+        Language::Cxx => Some("c++"),
+        Language::CxxHeader => Some("c++-header"),
+        Language::ObjectiveC => Some("objective-c"),
+        Language::ObjectiveCxx => Some("objective-c++"),
+        Language::Cuda => Some("cu"),
+        Language::Rust => None,          // Let the compiler decide
+        Language::GenericHeader => None, // Let the compiler decide
+    }
+}
+
 #[allow(clippy::too_many_arguments)]
 fn preprocess_cmd<T>(
     cmd: &mut T,
@@ -592,14 +661,11 @@ fn preprocess_cmd<T>(
 ) where
     T: RunCommand,
 {
-    let language = match parsed_args.language {
-        Language::C => "c",
-        Language::Cxx => "c++",
-        Language::ObjectiveC => "objective-c",
-        Language::ObjectiveCxx => "objective-c++",
-        Language::Cuda => "cu",
-    };
-    cmd.arg("-x").arg(language).arg("-E");
+    let language = language_to_gcc_arg(parsed_args.language);
+    if let Some(lang) = &language {
+        cmd.arg("-x").arg(lang);
+    }
+    cmd.arg("-E");
     // When performing distributed compilation, line number info is important for error
     // reporting and to not cause spurious compilation failure (e.g. no exceptions build
     // fails due to exceptions transitively included in the stdlib).
@@ -639,13 +705,25 @@ fn preprocess_cmd<T>(
         })
         .collect::<Vec<OsString>>();
 
-    cmd.arg(&parsed_args.input)
-        .args(&parsed_args.preprocessor_args)
+    let mut arch_args_to_use = &rewritten_arch_args;
+    let mut unique_rewritten = rewritten_arch_args.clone();
+    unique_rewritten.sort();
+    unique_rewritten.dedup();
+    if unique_rewritten.len() <= 1 {
+        // don't use rewritten arch args if there is only one arch
+        arch_args_to_use = &parsed_args.arch_args;
+    }
+
+    cmd.args(&parsed_args.preprocessor_args)
         .args(&parsed_args.dependency_args)
         .args(&parsed_args.common_args)
-        .args(&rewritten_arch_args)
+        .args(arch_args_to_use);
+    if parsed_args.double_dash_input {
+        cmd.arg("--");
+    }
+    cmd.arg(&parsed_args.input)
         .env_clear()
-        .envs(env_vars.iter().map(|&(ref k, ref v)| (k, v)))
+        .envs(env_vars.iter().map(|(k, v)| (k, v)))
         .current_dir(cwd);
     debug!("cmd after -arch rewrite: {:?}", cmd);
 }
@@ -709,24 +787,24 @@ pub fn generate_compile_commands(
 
     // Pass the language explicitly as we might have gotten it from the
     // command line.
-    let language = match parsed_args.language {
-        Language::C => "c",
-        Language::Cxx => "c++",
-        Language::ObjectiveC => "objective-c",
-        Language::ObjectiveCxx => "objective-c++",
-        Language::Cuda => "cu",
-    };
-    let mut arguments: Vec<OsString> = vec![
-        "-x".into(),
-        language.into(),
+    let language = language_to_gcc_arg(parsed_args.language);
+    let mut arguments: Vec<OsString> = vec![];
+    if let Some(lang) = &language {
+        arguments.extend(vec!["-x".into(), lang.into()])
+    }
+    arguments.extend(vec![
         parsed_args.compilation_flag.clone(),
-        parsed_args.input.clone().into(),
         "-o".into(),
         out_file.into(),
-    ];
-    arguments.extend(parsed_args.preprocessor_args.clone());
-    arguments.extend(parsed_args.common_args.clone());
-    arguments.extend(parsed_args.arch_args.clone());
+    ]);
+    arguments.extend_from_slice(&parsed_args.preprocessor_args);
+    arguments.extend_from_slice(&parsed_args.unhashed_args);
+    arguments.extend_from_slice(&parsed_args.common_args);
+    arguments.extend_from_slice(&parsed_args.arch_args);
+    if parsed_args.double_dash_input {
+        arguments.push("--".into());
+    }
+    arguments.push(parsed_args.input.clone().into());
     let command = CompileCommand {
         executable: executable.to_owned(),
         arguments,
@@ -739,28 +817,27 @@ pub fn generate_compile_commands(
     #[cfg(feature = "dist-client")]
     let dist_command = (|| {
         // https://gcc.gnu.org/onlinedocs/gcc-4.9.0/gcc/Overall-Options.html
-        let mut language: String = match parsed_args.language {
-            Language::C => "c",
-            Language::Cxx => "c++",
-            Language::ObjectiveC => "objective-c",
-            Language::ObjectiveCxx => "objective-c++",
-            Language::Cuda => "cu",
-        }
-        .into();
+        let mut language: Option<String> =
+            language_to_gcc_arg(parsed_args.language).map(|lang| lang.into());
         if !rewrite_includes_only {
             match parsed_args.language {
-                Language::C => language = "cpp-output".into(),
-                _ => language.push_str("-cpp-output"),
+                Language::C => language = Some("cpp-output".into()),
+                Language::GenericHeader | Language::CHeader | Language::CxxHeader => {}
+                _ => language.as_mut()?.push_str("-cpp-output"),
             }
         }
-        let mut arguments: Vec<String> = vec![
-            "-x".into(),
-            language,
+
+        let mut arguments: Vec<String> = vec![];
+        // Language needs to be before input
+        if let Some(lang) = &language {
+            arguments.extend(vec!["-x".into(), lang.into()])
+        }
+        arguments.extend(vec![
             parsed_args.compilation_flag.clone().into_string().ok()?,
             path_transformer.as_dist(&parsed_args.input)?,
             "-o".into(),
             path_transformer.as_dist(out_file)?,
-        ];
+        ]);
         if let CCompilerKind::Gcc = kind {
             // From https://gcc.gnu.org/onlinedocs/gcc/Preprocessor-Options.html:
             //
@@ -816,7 +893,7 @@ impl<'a> Iterator for ExpandIncludeFile<'a> {
                 None => return None,
             };
             let file = match arg.split_prefix("@") {
-                Some(arg) => self.cwd.join(&arg),
+                Some(arg) => self.cwd.join(arg),
                 None => return Some(arg),
             };
 
@@ -863,7 +940,7 @@ impl<'a> Iterator for ExpandIncludeFile<'a> {
 
 #[cfg(test)]
 mod test {
-    use std::fs::File;
+    use fs::File;
     use std::io::Write;
 
     use super::*;
@@ -871,6 +948,8 @@ mod test {
     use crate::mock_command::*;
     use crate::test::utils::*;
 
+    use temp_env::{with_var, with_var_unset};
+
     fn parse_arguments_(
         arguments: Vec<String>,
         plusplus: bool,
@@ -879,6 +958,20 @@ mod test {
         parse_arguments(&args, ".".as_ref(), &ARGS[..], plusplus, CCompilerKind::Gcc)
     }
 
+    fn parse_arguments_clang(
+        arguments: Vec<String>,
+        plusplus: bool,
+    ) -> CompilerArguments<ParsedArguments> {
+        let args = arguments.iter().map(OsString::from).collect::<Vec<_>>();
+        parse_arguments(
+            &args,
+            ".".as_ref(),
+            &ARGS[..],
+            plusplus,
+            CCompilerKind::Clang,
+        )
+    }
+
     #[test]
     fn test_parse_arguments_simple() {
         let args = stringvec!["-c", "foo.c", "-o", "foo.o"];
@@ -1274,6 +1367,65 @@ mod test {
         assert!(!msvc_show_includes);
     }
 
+    #[test]
+    fn test_parse_arguments_double_dash() {
+        let args = stringvec!["-c", "-o", "foo.o", "--", "foo.c"];
+        let ParsedArguments {
+            input,
+            double_dash_input,
+            common_args,
+            ..
+        } = match parse_arguments_(args.clone(), false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert_eq!(Some("foo.c"), input.to_str());
+        // GCC doesn't support double dashes. If we got one, we'll pass them
+        // through to GCC for it to error out.
+        assert!(!double_dash_input);
+        assert_eq!(ovec!["--"], common_args);
+
+        let ParsedArguments {
+            input,
+            double_dash_input,
+            common_args,
+            ..
+        } = match parse_arguments_clang(args, false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert_eq!(Some("foo.c"), input.to_str());
+        assert!(double_dash_input);
+        assert!(common_args.is_empty());
+
+        let args = stringvec!["-c", "-o", "foo.o", "foo.c", "--"];
+        let ParsedArguments {
+            input,
+            double_dash_input,
+            common_args,
+            ..
+        } = match parse_arguments_clang(args, false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert_eq!(Some("foo.c"), input.to_str());
+        // Double dash after input file is ignored.
+        assert!(!double_dash_input);
+        assert!(common_args.is_empty());
+
+        let args = stringvec!["-c", "-o", "foo.o", "foo.c", "--", "bar.c"];
+        assert_eq!(
+            CompilerArguments::CannotCache("multiple input files", Some("[\"bar.c\"]".to_string())),
+            parse_arguments_clang(args, false)
+        );
+
+        let args = stringvec!["-c", "-o", "foo.o", "foo.c", "--", "-fPIC"];
+        assert_eq!(
+            CompilerArguments::CannotCache("multiple input files", Some("[\"-fPIC\"]".to_string())),
+            parse_arguments_clang(args, false)
+        );
+    }
+
     #[test]
     fn test_parse_arguments_explicit_dep_target() {
         let args =
@@ -1415,7 +1567,43 @@ mod test {
 
     #[test]
     fn test_preprocess_cmd_rewrites_archs() {
-        let args = stringvec!["-arch", "arm64", "-arch", "i386", "-c", "foo.cc"];
+        with_var("SCCACHE_CACHE_MULTIARCH", Some("1"), || {
+            let args = stringvec!["-arch", "arm64", "-arch", "i386", "-c", "foo.cc"];
+            let parsed_args = match parse_arguments_(args, false) {
+                CompilerArguments::Ok(args) => args,
+                o => panic!("Got unexpected parse result: {:?}", o),
+            };
+            let mut cmd = MockCommand {
+                child: None,
+                args: vec![],
+            };
+            preprocess_cmd(
+                &mut cmd,
+                &parsed_args,
+                Path::new(""),
+                &[],
+                true,
+                CCompilerKind::Gcc,
+                true,
+                vec![],
+            );
+            // make sure the architectures were rewritten to prepocessor defines
+            let expected_args = ovec![
+                "-x",
+                "c++",
+                "-E",
+                "-fdirectives-only",
+                "-D__arm64__=1",
+                "-D__i386__=1",
+                "foo.cc"
+            ];
+            assert_eq!(cmd.args, expected_args);
+        });
+    }
+
+    #[test]
+    fn test_preprocess_cmd_doesnt_rewrite_single_arch() {
+        let args = stringvec!["-arch", "arm64", "-c", "foo.cc"];
         let parsed_args = match parse_arguments_(args, false) {
             CompilerArguments::Ok(args) => args,
             o => panic!("Got unexpected parse result: {:?}", o),
@@ -1440,13 +1628,38 @@ mod test {
             "c++",
             "-E",
             "-fdirectives-only",
-            "foo.cc",
-            "-D__arm64__=1",
-            "-D__i386__=1"
+            "-arch",
+            "arm64",
+            "foo.cc"
         ];
         assert_eq!(cmd.args, expected_args);
     }
 
+    #[test]
+    fn test_preprocess_double_dash_input() {
+        let args = stringvec!["-c", "-o", "foo.o", "--", "foo.c"];
+        let parsed_args = match parse_arguments_clang(args, false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        let mut cmd = MockCommand {
+            child: None,
+            args: vec![],
+        };
+        preprocess_cmd(
+            &mut cmd,
+            &parsed_args,
+            Path::new(""),
+            &[],
+            true,
+            CCompilerKind::Clang,
+            true,
+            vec![],
+        );
+        let expected_args = ovec!["-x", "c", "-E", "-frewrite-includes", "--", "foo.c"];
+        assert_eq!(cmd.args, expected_args);
+    }
+
     #[test]
     fn pedantic_default() {
         let args = stringvec!["-pedantic", "-c", "foo.cc"];
@@ -1668,6 +1881,24 @@ mod test {
         );
     }
 
+    #[test]
+    fn test_parse_arguments_multiarch_cache_disabled() {
+        with_var_unset("SCCACHE_CACHE_MULTIARCH", || {
+            assert_eq!(
+                CompilerArguments::CannotCache(
+                    "multiple different -arch, and SCCACHE_CACHE_MULTIARCH not set",
+                    None
+                ),
+                parse_arguments_(
+                    stringvec![
+                        "-fPIC", "-arch", "arm64", "-arch", "i386", "-o", "foo.o", "-c", "foo.cpp"
+                    ],
+                    false
+                )
+            )
+        });
+    }
+
     #[test]
     fn test_parse_arguments_multiple_arch() {
         match parse_arguments_(
@@ -1678,47 +1909,50 @@ mod test {
             o => panic!("Got unexpected parse result: {:?}", o),
         }
 
-        match parse_arguments_(
-            stringvec!["-arch", "arm64", "-arch", "arm64", "-o", "foo.o", "-c", "foo.cpp"],
-            false,
-        ) {
-            CompilerArguments::Ok(_) => {}
-            o => panic!("Got unexpected parse result: {:?}", o),
-        }
+        with_var("SCCACHE_CACHE_MULTIARCH", Some("1"), || {
+            match parse_arguments_(
+                stringvec!["-arch", "arm64", "-arch", "arm64", "-o", "foo.o", "-c", "foo.cpp"],
+                false,
+            ) {
+                CompilerArguments::Ok(_) => {}
+                o => panic!("Got unexpected parse result: {:?}", o),
+            }
 
-        let args =
-            stringvec!["-fPIC", "-arch", "arm64", "-arch", "i386", "-o", "foo.o", "-c", "foo.cpp"];
-        let ParsedArguments {
-            input,
-            language,
-            compilation_flag,
-            outputs,
-            preprocessor_args,
-            msvc_show_includes,
-            common_args,
-            arch_args,
-            ..
-        } = match parse_arguments_(args, false) {
-            CompilerArguments::Ok(args) => args,
-            o => panic!("Got unexpected parse result: {:?}", o),
-        };
-        assert_eq!(Some("foo.cpp"), input.to_str());
-        assert_eq!(Language::Cxx, language);
-        assert_eq!(Some("-c"), compilation_flag.to_str());
-        assert_map_contains!(
-            outputs,
-            (
-                "obj",
-                ArtifactDescriptor {
-                    path: "foo.o".into(),
-                    optional: false
-                }
-            )
-        );
-        assert!(preprocessor_args.is_empty());
-        assert_eq!(ovec!["-fPIC"], common_args);
-        assert_eq!(ovec!["-arch", "arm64", "-arch", "i386"], arch_args);
-        assert!(!msvc_show_includes);
+            let args = stringvec![
+                "-fPIC", "-arch", "arm64", "-arch", "i386", "-o", "foo.o", "-c", "foo.cpp"
+            ];
+            let ParsedArguments {
+                input,
+                language,
+                compilation_flag,
+                outputs,
+                preprocessor_args,
+                msvc_show_includes,
+                common_args,
+                arch_args,
+                ..
+            } = match parse_arguments_(args, false) {
+                CompilerArguments::Ok(args) => args,
+                o => panic!("Got unexpected parse result: {:?}", o),
+            };
+            assert_eq!(Some("foo.cpp"), input.to_str());
+            assert_eq!(Language::Cxx, language);
+            assert_eq!(Some("-c"), compilation_flag.to_str());
+            assert_map_contains!(
+                outputs,
+                (
+                    "obj",
+                    ArtifactDescriptor {
+                        path: "foo.o".into(),
+                        optional: false
+                    }
+                )
+            );
+            assert!(preprocessor_args.is_empty());
+            assert_eq!(ovec!["-fPIC"], common_args);
+            assert_eq!(ovec!["-arch", "arm64", "-arch", "i386"], arch_args);
+            assert!(!msvc_show_includes);
+        });
     }
 
     #[test]
@@ -1771,6 +2005,7 @@ mod test {
         let f = TestFixture::new();
         let parsed_args = ParsedArguments {
             input: "foo.c".into(),
+            double_dash_input: false,
             language: Language::C,
             compilation_flag: "-c".into(),
             depfile: None,
@@ -1787,11 +2022,13 @@ mod test {
             preprocessor_args: vec![],
             common_args: vec![],
             arch_args: vec![],
+            unhashed_args: vec![],
             extra_hash_files: vec![],
             msvc_show_includes: false,
             profile_generate: false,
             color_mode: ColorMode::Auto,
             suppress_rewrite_includes_only: false,
+            too_hard_for_preprocessor_cache_mode: false,
         };
         let compiler = &f.bins[0];
         // Compiler invocation.
@@ -1817,6 +2054,30 @@ mod test {
         assert_eq!(0, creator.lock().unwrap().children.len());
     }
 
+    #[test]
+    fn test_compile_double_dash_input() {
+        let args = stringvec!["-c", "-o", "foo.o", "--", "foo.c"];
+        let parsed_args = match parse_arguments_clang(args, false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        let f = TestFixture::new();
+        let compiler = &f.bins[0];
+        let mut path_transformer = dist::PathTransformer::default();
+        let (command, _, _) = generate_compile_commands(
+            &mut path_transformer,
+            compiler,
+            &parsed_args,
+            f.tempdir.path(),
+            &[],
+            CCompilerKind::Clang,
+            false,
+        )
+        .unwrap();
+        let expected_args = ovec!["-x", "c", "-c", "-o", "foo.o", "--", "foo.c"];
+        assert_eq!(command.arguments, expected_args);
+    }
+
     #[test]
     fn test_parse_arguments_plusplus() {
         let args = stringvec!["-c", "foo.c", "-o", "foo.o"];
@@ -1850,4 +2111,100 @@ mod test {
         assert!(common_args.is_empty());
         assert!(!msvc_show_includes);
     }
+
+    #[test]
+    fn test_pch_explicit() {
+        let args = stringvec!["-c", "-x", "c++-header", "pch.h", "-o", "pch.pch"];
+        let parsed_args = match parse_arguments_(args, false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        let mut cmd = MockCommand {
+            child: None,
+            args: vec![],
+        };
+        preprocess_cmd(
+            &mut cmd,
+            &parsed_args,
+            Path::new(""),
+            &[],
+            true,
+            CCompilerKind::Gcc,
+            true,
+            vec![],
+        );
+        assert!(cmd.args.contains(&"-x".into()) && cmd.args.contains(&"c++-header".into()));
+    }
+
+    #[test]
+    fn test_pch_implicit() {
+        let args = stringvec!["-c", "pch.hpp", "-o", "pch.pch"];
+        let parsed_args = match parse_arguments_(args, false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        let mut cmd = MockCommand {
+            child: None,
+            args: vec![],
+        };
+        preprocess_cmd(
+            &mut cmd,
+            &parsed_args,
+            Path::new(""),
+            &[],
+            true,
+            CCompilerKind::Gcc,
+            true,
+            vec![],
+        );
+        assert!(cmd.args.contains(&"-x".into()) && cmd.args.contains(&"c++-header".into()));
+    }
+
+    #[test]
+    fn test_pch_generic() {
+        let args = stringvec!["-c", "pch.h", "-o", "pch.pch"];
+        let parsed_args = match parse_arguments_(args, false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        let mut cmd = MockCommand {
+            child: None,
+            args: vec![],
+        };
+        preprocess_cmd(
+            &mut cmd,
+            &parsed_args,
+            Path::new(""),
+            &[],
+            true,
+            CCompilerKind::Gcc,
+            true,
+            vec![],
+        );
+        assert!(!cmd.args.contains(&"-x".into()));
+    }
+
+    #[test]
+    fn test_too_hard_for_preprocessor_cache_mode() {
+        let args = stringvec!["-c", "foo.c", "-o", "foo.o"];
+        let parsed_args = match parse_arguments_(args, false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert!(!parsed_args.too_hard_for_preprocessor_cache_mode);
+
+        let args = stringvec!["-c", "foo.c", "-o", "foo.o", "-Xpreprocessor", "-M"];
+        let parsed_args = match parse_arguments_(args, false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert!(parsed_args.too_hard_for_preprocessor_cache_mode);
+
+        let args = stringvec!["-c", "foo.c", "-o", "foo.o", r#"-Wp,-DFOO="something""#];
+        let parsed_args = match parse_arguments_(args, false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert!(parsed_args.too_hard_for_preprocessor_cache_mode);
+    }
 }
