@@ -12,6 +12,7 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+use crate::cache::CacheMode;
 use directories::ProjectDirs;
 use fs::File;
 use fs_err as fs;
@@ -33,10 +34,11 @@ use std::result::Result as StdResult;
 use std::str::FromStr;
 use std::sync::Mutex;
 
+pub use crate::cache::PreprocessorCacheModeConfig;
 use crate::errors::*;
 
 static CACHED_CONFIG_PATH: Lazy<PathBuf> = Lazy::new(CachedConfig::file_config_path);
-static CACHED_CONFIG: Lazy<Mutex<Option<CachedFileConfig>>> = Lazy::new(|| Mutex::new(None));
+static CACHED_CONFIG: Mutex<Option<CachedFileConfig>> = Mutex::new(None);
 
 const ORGANIZATION: &str = "Mozilla";
 const APP_NAME: &str = "sccache";
@@ -160,6 +162,8 @@ pub struct DiskCacheConfig {
     pub dir: PathBuf,
     // TODO: use deserialize_with to allow human-readable sizes in toml
     pub size: u64,
+    pub preprocessor_cache_mode: PreprocessorCacheModeConfig,
+    pub rw_mode: CacheModeConfig,
 }
 
 impl Default for DiskCacheConfig {
@@ -167,19 +171,30 @@ impl Default for DiskCacheConfig {
         DiskCacheConfig {
             dir: default_disk_cache_dir(),
             size: default_disk_cache_size(),
+            preprocessor_cache_mode: PreprocessorCacheModeConfig::activated(),
+            rw_mode: CacheModeConfig::ReadWrite,
         }
     }
 }
 
 #[derive(Copy, Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
-pub enum GCSCacheRWMode {
+pub enum CacheModeConfig {
     #[serde(rename = "READ_ONLY")]
     ReadOnly,
     #[serde(rename = "READ_WRITE")]
     ReadWrite,
 }
 
+impl From<CacheModeConfig> for CacheMode {
+    fn from(value: CacheModeConfig) -> Self {
+        match value {
+            CacheModeConfig::ReadOnly => CacheMode::ReadOnly,
+            CacheModeConfig::ReadWrite => CacheMode::ReadWrite,
+        }
+    }
+}
+
 #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
 pub struct GCSCacheConfig {
@@ -187,7 +202,7 @@ pub struct GCSCacheConfig {
     pub key_prefix: String,
     pub cred_path: Option<String>,
     pub service_account: Option<String>,
-    pub rw_mode: GCSCacheRWMode,
+    pub rw_mode: CacheModeConfig,
     pub credential_url: Option<String>,
 }
 
@@ -220,10 +235,21 @@ pub struct MemcachedCacheConfig {
     pub expiration: u32,
 }
 
+/// redis has no default TTL - all caches live forever
+///
+/// We keep the TTL as 0 here as redis does
+///
+/// Please change this value freely if we have a better choice.
+const DEFAULT_REDIS_CACHE_TTL: u64 = 0;
 #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
 pub struct RedisCacheConfig {
     pub url: String,
+    /// the ttl time in seconds.
+    ///
+    /// Default to infinity (0)
+    #[serde(default)]
+    pub ttl: u64,
 }
 
 #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
@@ -245,6 +271,16 @@ pub struct S3CacheConfig {
     pub no_credentials: bool,
     pub endpoint: Option<String>,
     pub use_ssl: Option<bool>,
+    pub server_side_encryption: Option<bool>,
+}
+
+#[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
+#[serde(deny_unknown_fields)]
+pub struct OSSCacheConfig {
+    pub bucket: String,
+    pub key_prefix: String,
+    pub endpoint: Option<String>,
+    pub no_credentials: bool,
 }
 
 #[derive(Debug, PartialEq, Eq)]
@@ -256,6 +292,7 @@ pub enum CacheType {
     Redis(RedisCacheConfig),
     S3(S3CacheConfig),
     Webdav(WebdavCacheConfig),
+    OSS(OSSCacheConfig),
 }
 
 #[derive(Debug, Default, Serialize, Deserialize, PartialEq, Eq)]
@@ -269,6 +306,7 @@ pub struct CacheConfigs {
     pub redis: Option<RedisCacheConfig>,
     pub s3: Option<S3CacheConfig>,
     pub webdav: Option<WebdavCacheConfig>,
+    pub oss: Option<OSSCacheConfig>,
 }
 
 impl CacheConfigs {
@@ -284,6 +322,7 @@ impl CacheConfigs {
             redis,
             s3,
             webdav,
+            oss,
         } = self;
 
         let cache_type = s3
@@ -293,7 +332,8 @@ impl CacheConfigs {
             .or_else(|| gcs.map(CacheType::GCS))
             .or_else(|| gha.map(CacheType::GHA))
             .or_else(|| azure.map(CacheType::Azure))
-            .or_else(|| webdav.map(CacheType::Webdav));
+            .or_else(|| webdav.map(CacheType::Webdav))
+            .or_else(|| oss.map(CacheType::OSS));
 
         let fallback = disk.unwrap_or_default();
 
@@ -311,6 +351,7 @@ impl CacheConfigs {
             redis,
             s3,
             webdav,
+            oss,
         } = other;
 
         if azure.is_some() {
@@ -337,6 +378,10 @@ impl CacheConfigs {
         if webdav.is_some() {
             self.webdav = webdav
         }
+
+        if oss.is_some() {
+            self.oss = oss
+        }
     }
 }
 
@@ -519,6 +564,16 @@ fn config_from_env() -> Result<EnvConfig> {
         let use_ssl = env::var("SCCACHE_S3_USE_SSL")
             .ok()
             .map(|value| value != "off");
+        let server_side_encryption =
+            env::var("SCCACHE_S3_SERVER_SIDE_ENCRYPTION")
+                .ok()
+                .map_or(Ok(Some(false)), |val| match val.as_str() {
+                    "true" | "1" => Ok(Some(true)),
+                    "false" | "0" => Ok(Some(false)),
+                    _ => bail!(
+                        "SCCACHE_S3_SERVER_SIDE_ENCRYPTION must be 'true', '1', 'false', or '0'."
+                    ),
+                })?;
         let endpoint = env::var("SCCACHE_ENDPOINT").ok();
         let key_prefix = env::var("SCCACHE_S3_KEY_PREFIX")
             .ok()
@@ -535,6 +590,7 @@ fn config_from_env() -> Result<EnvConfig> {
             key_prefix,
             endpoint,
             use_ssl,
+            server_side_encryption,
         })
     } else {
         None
@@ -548,9 +604,15 @@ fn config_from_env() -> Result<EnvConfig> {
     }
 
     // ======= redis =======
+    let ttl = match env::var("SCCACHE_REDIS_TTL") {
+        Ok(v) => v
+            .parse()
+            .map_err(|err| anyhow!("SCCACHE_REDIS_TTL value is invalid: {err:?}"))?,
+        Err(_) => DEFAULT_REDIS_CACHE_TTL,
+    };
     let redis = env::var("SCCACHE_REDIS")
         .ok()
-        .map(|url| RedisCacheConfig { url });
+        .map(|url| RedisCacheConfig { url, ttl });
 
     // ======= memcached =======
     let expiration = match env::var("SCCACHE_MEMCACHED_EXPIRATION").ok() {
@@ -588,7 +650,7 @@ fn config_from_env() -> Result<EnvConfig> {
 
         if env::var("SCCACHE_GCS_OAUTH_URL").is_ok() {
             eprintln!("SCCACHE_GCS_OAUTH_URL has been deprecated");
-            eprintln!("if you intend to use vm metadata for auth, please set correct service account intead");
+            eprintln!("if you intend to use vm metadata for auth, please set correct service account instead");
         }
 
         let credential_url = env::var("SCCACHE_GCS_CREDENTIALS_URL").ok();
@@ -597,17 +659,17 @@ fn config_from_env() -> Result<EnvConfig> {
         let service_account = env::var("SCCACHE_GCS_SERVICE_ACCOUNT").ok();
 
         let rw_mode = match env::var("SCCACHE_GCS_RW_MODE").as_ref().map(String::as_str) {
-            Ok("READ_ONLY") => GCSCacheRWMode::ReadOnly,
-            Ok("READ_WRITE") => GCSCacheRWMode::ReadWrite,
+            Ok("READ_ONLY") => CacheModeConfig::ReadOnly,
+            Ok("READ_WRITE") => CacheModeConfig::ReadWrite,
             // TODO: unsure if these should warn during the configuration loading
             // or at the time when they're actually used to connect to GCS
             Ok(_) => {
                 warn!("Invalid SCCACHE_GCS_RW_MODE-- defaulting to READ_ONLY.");
-                GCSCacheRWMode::ReadOnly
+                CacheModeConfig::ReadOnly
             }
             _ => {
                 warn!("No SCCACHE_GCS_RW_MODE specified-- defaulting to READ_ONLY.");
-                GCSCacheRWMode::ReadOnly
+                CacheModeConfig::ReadOnly
             }
         };
 
@@ -689,16 +751,86 @@ fn config_from_env() -> Result<EnvConfig> {
         None
     };
 
+    // ======= OSS =======
+    let oss = if let Ok(bucket) = env::var("SCCACHE_OSS_BUCKET") {
+        let endpoint = env::var("SCCACHE_OSS_ENDPOINT").ok();
+        let key_prefix = env::var("SCCACHE_OSS_KEY_PREFIX")
+            .ok()
+            .as_ref()
+            .map(|s| s.trim_end_matches('/'))
+            .filter(|s| !s.is_empty())
+            .unwrap_or_default()
+            .to_owned();
+
+        let no_credentials =
+            env::var("SCCACHE_OSS_NO_CREDENTIALS").map_or(Ok(false), |val| match val.as_str() {
+                "true" | "1" => Ok(true),
+                "false" | "0" => Ok(false),
+                _ => bail!("SCCACHE_OSS_NO_CREDENTIALS must be 'true', '1', 'false', or '0'."),
+            })?;
+
+        Some(OSSCacheConfig {
+            bucket,
+            endpoint,
+            key_prefix,
+            no_credentials,
+        })
+    } else {
+        None
+    };
+
+    if oss
+        .as_ref()
+        .map(|oss| oss.no_credentials)
+        .unwrap_or_default()
+        && (env::var_os("ALIBABA_CLOUD_ACCESS_KEY_ID").is_some()
+            || env::var_os("ALIBABA_CLOUD_ACCESS_KEY_SECRET").is_some())
+    {
+        bail!("If setting OSS credentials, SCCACHE_OSS_NO_CREDENTIALS must not be set.");
+    }
+
     // ======= Local =======
     let disk_dir = env::var_os("SCCACHE_DIR").map(PathBuf::from);
     let disk_sz = env::var("SCCACHE_CACHE_SIZE")
         .ok()
         .and_then(|v| parse_size(&v));
 
-    let disk = if disk_dir.is_some() || disk_sz.is_some() {
+    let mut preprocessor_mode_config = PreprocessorCacheModeConfig::activated();
+    let preprocessor_mode_overridden = match env::var("SCCACHE_DIRECT").as_deref() {
+        Ok("on") | Ok("true") => {
+            preprocessor_mode_config.use_preprocessor_cache_mode = true;
+            true
+        }
+        Ok("off") | Ok("false") => {
+            preprocessor_mode_config.use_preprocessor_cache_mode = false;
+            true
+        }
+        _ => false,
+    };
+
+    let (disk_rw_mode, disk_rw_mode_overridden) = match env::var("SCCACHE_LOCAL_RW_MODE")
+        .as_ref()
+        .map(String::as_str)
+    {
+        Ok("READ_ONLY") => (CacheModeConfig::ReadOnly, true),
+        Ok("READ_WRITE") => (CacheModeConfig::ReadWrite, true),
+        Ok(_) => {
+            warn!("Invalid SCCACHE_LOCAL_RW_MODE-- defaulting to READ_WRITE.");
+            (CacheModeConfig::ReadWrite, false)
+        }
+        _ => (CacheModeConfig::ReadWrite, false),
+    };
+
+    let any_overridden = disk_dir.is_some()
+        || disk_sz.is_some()
+        || preprocessor_mode_overridden
+        || disk_rw_mode_overridden;
+    let disk = if any_overridden {
         Some(DiskCacheConfig {
             dir: disk_dir.unwrap_or_else(default_disk_cache_dir),
             size: disk_sz.unwrap_or_else(default_disk_cache_size),
+            preprocessor_cache_mode: preprocessor_mode_config,
+            rw_mode: disk_rw_mode,
         })
     } else {
         None
@@ -713,6 +845,7 @@ fn config_from_env() -> Result<EnvConfig> {
         redis,
         s3,
         webdav,
+        oss,
     };
 
     Ok(EnvConfig { cache })
@@ -1041,9 +1174,12 @@ fn config_overrides() {
             disk: Some(DiskCacheConfig {
                 dir: "/env-cache".into(),
                 size: 5,
+                preprocessor_cache_mode: Default::default(),
+                rw_mode: CacheModeConfig::ReadWrite,
             }),
             redis: Some(RedisCacheConfig {
                 url: "myotherredisurl".to_owned(),
+                ttl: 24 * 3600,
             }),
             ..Default::default()
         },
@@ -1054,6 +1190,8 @@ fn config_overrides() {
             disk: Some(DiskCacheConfig {
                 dir: "/file-cache".into(),
                 size: 15,
+                preprocessor_cache_mode: Default::default(),
+                rw_mode: CacheModeConfig::ReadWrite,
             }),
             memcached: Some(MemcachedCacheConfig {
                 url: "memurl".to_owned(),
@@ -1061,6 +1199,7 @@ fn config_overrides() {
             }),
             redis: Some(RedisCacheConfig {
                 url: "myredisurl".to_owned(),
+                ttl: 24 * 3600,
             }),
             ..Default::default()
         },
@@ -1072,11 +1211,14 @@ fn config_overrides() {
         Config::from_env_and_file_configs(env_conf, file_conf),
         Config {
             cache: Some(CacheType::Redis(RedisCacheConfig {
-                url: "myotherredisurl".to_owned()
+                url: "myotherredisurl".to_owned(),
+                ttl: 24 * 3600,
             }),),
             fallback_cache: DiskCacheConfig {
                 dir: "/env-cache".into(),
                 size: 5,
+                preprocessor_cache_mode: Default::default(),
+                rw_mode: CacheModeConfig::ReadWrite,
             },
             dist: Default::default(),
             server_startup_timeout: None,
@@ -1182,7 +1324,7 @@ fn test_gcs_service_account() {
         }) => {
             assert_eq!(bucket, "my-bucket");
             assert_eq!(service_account, Some("my@example.com".to_string()));
-            assert_eq!(rw_mode, GCSCacheRWMode::ReadWrite);
+            assert_eq!(rw_mode, CacheModeConfig::ReadWrite);
         }
         None => unreachable!(),
     };
@@ -1236,6 +1378,7 @@ expiration = 86400
 
 [cache.redis]
 url = "redis://user:passwd@1.2.3.4:6379/1"
+ttl = 86400
 
 [cache.s3]
 bucket = "name"
@@ -1244,6 +1387,7 @@ endpoint = "s3-us-east-1.amazonaws.com"
 use_ssl = true
 key_prefix = "s3prefix"
 no_credentials = true
+server_side_encryption = false
 
 [cache.webdav]
 endpoint = "http://127.0.0.1:8080"
@@ -1251,6 +1395,12 @@ key_prefix = "webdavprefix"
 username = "webdavusername"
 password = "webdavpassword"
 token = "webdavtoken"
+
+[cache.oss]
+bucket = "name"
+endpoint = "oss-us-east-1.aliyuncs.com"
+key_prefix = "ossprefix"
+no_credentials = true
 "#;
 
     let file_config: FileConfig = toml::from_str(CONFIG_STR).expect("Is valid toml.");
@@ -1262,12 +1412,14 @@ token = "webdavtoken"
                 disk: Some(DiskCacheConfig {
                     dir: PathBuf::from("/tmp/.cache/sccache"),
                     size: 7 * 1024 * 1024 * 1024,
+                    preprocessor_cache_mode: PreprocessorCacheModeConfig::activated(),
+                    rw_mode: CacheModeConfig::ReadWrite,
                 }),
                 gcs: Some(GCSCacheConfig {
                     bucket: "bucket".to_owned(),
                     cred_path: Some("/psst/secret/cred".to_string()),
                     service_account: Some("example_service_account".to_string()),
-                    rw_mode: GCSCacheRWMode::ReadOnly,
+                    rw_mode: CacheModeConfig::ReadOnly,
                     key_prefix: "prefix".into(),
                     credential_url: None,
                 }),
@@ -1277,6 +1429,7 @@ token = "webdavtoken"
                 }),
                 redis: Some(RedisCacheConfig {
                     url: "redis://user:passwd@1.2.3.4:6379/1".to_owned(),
+                    ttl: 24 * 3600,
                 }),
                 memcached: Some(MemcachedCacheConfig {
                     url: "...".to_owned(),
@@ -1289,6 +1442,7 @@ token = "webdavtoken"
                     use_ssl: Some(true),
                     key_prefix: "s3prefix".into(),
                     no_credentials: true,
+                    server_side_encryption: Some(false)
                 }),
                 webdav: Some(WebdavCacheConfig {
                     endpoint: "http://127.0.0.1:8080".to_string(),
@@ -1296,7 +1450,13 @@ token = "webdavtoken"
                     username: Some("webdavusername".to_string()),
                     password: Some("webdavpassword".to_string()),
                     token: Some("webdavtoken".to_string()),
-                })
+                }),
+                oss: Some(OSSCacheConfig {
+                    bucket: "name".to_owned(),
+                    endpoint: Some("oss-us-east-1.aliyuncs.com".to_owned()),
+                    key_prefix: "ossprefix".into(),
+                    no_credentials: true,
+                }),
             },
             dist: DistConfig {
                 auth: DistAuth::Token {
