@@ -15,7 +15,7 @@
 use crate::cache::{storage_from_config, Storage};
 use crate::compiler::{
     get_compiler_info, CacheControl, CompileResult, Compiler, CompilerArguments, CompilerHasher,
-    CompilerKind, CompilerProxy, DistType, MissType,
+    CompilerKind, CompilerProxy, DistType, Language, MissType,
 };
 #[cfg(feature = "dist-client")]
 use crate::config;
@@ -29,15 +29,17 @@ use crate::util;
 use anyhow::Context as _;
 use bytes::{buf::BufMut, Bytes, BytesMut};
 use filetime::FileTime;
+use fs::metadata;
+use fs_err as fs;
 use futures::channel::mpsc;
 use futures::future::FutureExt;
 use futures::{future, stream, Sink, SinkExt, Stream, StreamExt, TryFutureExt};
-use futures_locks::RwLock;
 use number_prefix::NumberPrefix;
+use serde::{Deserialize, Serialize};
+use std::cell::Cell;
 use std::collections::HashMap;
 use std::env;
 use std::ffi::{OsStr, OsString};
-use std::fs::metadata;
 use std::future::Future;
 use std::io::{self, Write};
 use std::marker::Unpin;
@@ -48,12 +50,13 @@ use std::path::PathBuf;
 use std::pin::Pin;
 use std::process::{ExitStatus, Output};
 use std::sync::Arc;
-use std::sync::Mutex;
 use std::task::{Context, Poll, Waker};
 use std::time::Duration;
 #[cfg(feature = "dist-client")]
 use std::time::Instant;
 use std::u64;
+use tokio::sync::Mutex;
+use tokio::sync::RwLock;
 use tokio::{
     io::{AsyncRead, AsyncWrite},
     net::TcpListener,
@@ -87,7 +90,7 @@ pub enum ServerStartup {
     Err { reason: String },
 }
 
-/// Get the time the server should idle for before shutting down.
+/// Get the time the server should idle for before shutting down, in seconds.
 fn get_idle_timeout() -> u64 {
     // A value of 0 disables idle shutdown entirely.
     env::var("SCCACHE_IDLE_TIMEOUT")
@@ -114,7 +117,7 @@ fn notify_server_startup(name: &Option<OsString>, status: ServerStartup) -> Resu
 
 #[cfg(windows)]
 fn notify_server_startup(name: &Option<OsString>, status: ServerStartup) -> Result<()> {
-    use std::fs::OpenOptions;
+    use fs::OpenOptions;
 
     let name = match *name {
         Some(ref s) => s,
@@ -223,8 +226,10 @@ impl DistClientContainer {
             | DistClientState::FailWithMessage(cfg, _)
             | DistClientState::RetryCreateAt(cfg, _) => {
                 warn!("State reset. Will recreate");
-                *state =
-                    DistClientState::RetryCreateAt(cfg, Instant::now() - Duration::from_secs(1));
+                *state = DistClientState::RetryCreateAt(
+                    cfg,
+                    Instant::now().checked_sub(Duration::from_secs(1)).unwrap(),
+                );
             }
             DistClientState::Disabled => (),
         }
@@ -275,8 +280,10 @@ impl DistClientContainer {
             };
             // The client is most likely mis-configured, make sure we
             // re-create on our next attempt.
-            *state =
-                DistClientState::RetryCreateAt(config, Instant::now() - Duration::from_secs(1));
+            *state = DistClientState::RetryCreateAt(
+                config,
+                Instant::now().checked_sub(Duration::from_secs(1)).unwrap(),
+            );
         }
         res
     }
@@ -386,12 +393,29 @@ impl DistClientContainer {
     }
 }
 
+thread_local! {
+    /// catch_unwind doesn't provide panic location, so we store that
+    /// information via a panic hook to be used when catch_unwind
+    /// catches a panic.
+    static PANIC_LOCATION: Cell<Option<(String, u32, u32)>> = Cell::new(None);
+}
+
 /// Start an sccache server, listening on `port`.
 ///
 /// Spins an event loop handling client connections until a client
 /// requests a shutdown.
 pub fn start_server(config: &Config, port: u16) -> Result<()> {
     info!("start_server: port: {}", port);
+    let panic_hook = std::panic::take_hook();
+    std::panic::set_hook(Box::new(move |info| {
+        PANIC_LOCATION.with(|l| {
+            l.set(
+                info.location()
+                    .map(|loc| (loc.file().to_string(), loc.line(), loc.column())),
+            )
+        });
+        panic_hook(info)
+    }));
     let client = unsafe { Client::new() };
     let runtime = tokio::runtime::Builder::new_multi_thread()
         .enable_all()
@@ -418,6 +442,25 @@ pub fn start_server(config: &Config, port: u16) -> Result<()> {
         }
     };
 
+    let cache_mode = runtime.block_on(async {
+        match storage.check().await {
+            Ok(mode) => Ok(mode),
+            Err(err) => {
+                error!("storage check failed for: {err:?}");
+
+                notify_server_startup(
+                    &notify,
+                    ServerStartup::Err {
+                        reason: err.to_string(),
+                    },
+                )?;
+
+                Err(err)
+            }
+        }
+    })?;
+    info!("server has setup with {cache_mode:?}");
+
     let res =
         SccacheServer::<ProcessCommandCreator>::new(port, runtime, client, dist_client, storage);
     match res {
@@ -566,6 +609,7 @@ impl<C: CommandCreatorSync> SccacheServer<C> {
 
                 // We're not interested if the task panicked; immediately process
                 // another connection
+                #[allow(clippy::let_underscore_future)]
                 let _ = tokio::spawn(conn);
             }
         };
@@ -664,7 +708,7 @@ where
     C: Send,
 {
     /// Server statistics.
-    stats: Arc<RwLock<ServerStats>>,
+    stats: Arc<Mutex<ServerStats>>,
 
     /// Distributed sccache client
     dist_client: Arc<DistClientContainer>,
@@ -741,7 +785,7 @@ where
             match req.into_inner() {
                 Request::Compile(compile) => {
                     debug!("handle_client: compile");
-                    me.stats.write().await.compile_requests += 1;
+                    me.stats.lock().await.compile_requests += 1;
                     me.handle_compile(compile).await
                 }
                 Request::GetStats => {
@@ -761,10 +805,7 @@ where
                 Request::ZeroStats => {
                     debug!("handle_client: zero_stats");
                     me.zero_stats().await;
-                    me.get_info()
-                        .await
-                        .map(|i| Response::Stats(Box::new(i)))
-                        .map(Message::WithoutBody)
+                    Ok(Message::WithoutBody(Response::ZeroStats))
                 }
                 Request::Shutdown => {
                     debug!("handle_client: shutdown");
@@ -806,11 +847,11 @@ where
         info: ActiveInfo,
     ) -> SccacheService<C> {
         SccacheService {
-            stats: Arc::new(RwLock::new(ServerStats::default())),
+            stats: Arc::default(),
             dist_client: Arc::new(dist_client),
             storage,
-            compilers: Arc::new(RwLock::new(HashMap::new())),
-            compiler_proxies: Arc::new(RwLock::new(HashMap::new())),
+            compilers: Arc::default(),
+            compiler_proxies: Arc::default(),
             rt,
             creator: C::new(client),
             tx,
@@ -868,21 +909,13 @@ where
 
     /// Get info and stats about the cache.
     async fn get_info(&self) -> Result<ServerInfo> {
-        let stats = self.stats.read().await.clone();
-        let cache_location = self.storage.location();
-        futures::try_join!(self.storage.current_size(), self.storage.max_size()).map(
-            move |(cache_size, max_cache_size)| ServerInfo {
-                stats,
-                cache_location,
-                cache_size,
-                max_cache_size,
-            },
-        )
+        let stats = self.stats.lock().await.clone();
+        ServerInfo::new(stats, Some(&*self.storage)).await
     }
 
     /// Zero stats about the cache.
     async fn zero_stats(&self) {
-        *self.stats.write().await = ServerStats::default();
+        *self.stats.lock().await = ServerStats::default();
     }
 
     /// Handle a compile request from a client.
@@ -983,7 +1016,7 @@ where
 
         let opt = match me1.compilers.read().await.get(&resolved_compiler_path) {
             // It's a hit only if the mtime and dist archive data matches.
-            Some(&Some(ref entry)) => {
+            Some(Some(entry)) => {
                 if entry.mtime == mtime && entry.dist_info == dist_info {
                     Some(entry.compiler.box_clone())
                 } else {
@@ -1071,11 +1104,10 @@ where
         cwd: PathBuf,
         env_vars: Vec<(OsString, OsString)>,
     ) -> SccacheResponse {
-        let mut stats = self.stats.write().await;
         match compiler {
             Err(e) => {
                 debug!("check_compiler: Unsupported compiler: {}", e.to_string());
-                stats.requests_unsupported_compiler += 1;
+                self.stats.lock().await.requests_unsupported_compiler += 1;
                 return Message::WithoutBody(Response::Compile(
                     CompileResponse::UnsupportedCompiler(OsString::from(e.to_string())),
                 ));
@@ -1087,7 +1119,7 @@ where
                 match c.parse_arguments(&cmd, &cwd, &env_vars) {
                     CompilerArguments::Ok(hasher) => {
                         debug!("parse_arguments: Ok: {:?}", cmd);
-                        stats.requests_executed += 1;
+                        self.stats.lock().await.requests_executed += 1;
                         let (tx, rx) = Body::pair();
                         self.start_compile_task(c, hasher, cmd, cwd, env_vars, tx);
                         let res = CompileResponse::CompileStarted;
@@ -1102,12 +1134,13 @@ where
                         } else {
                             debug!("parse_arguments: CannotCache({}): {:?}", why, cmd)
                         }
+                        let mut stats = self.stats.lock().await;
                         stats.requests_not_cacheable += 1;
                         *stats.not_cached.entry(why.to_string()).or_insert(0) += 1;
                     }
                     CompilerArguments::NotCompilation => {
                         debug!("parse_arguments: NotCompilation: {:?}", cmd);
-                        stats.requests_not_compile += 1;
+                        self.stats.lock().await.requests_not_compile += 1;
                     }
                 }
             }
@@ -1131,7 +1164,7 @@ where
     ) {
         let force_recache = env_vars
             .iter()
-            .any(|&(ref k, ref _v)| k.as_os_str() == OsStr::new("SCCACHE_RECACHE"));
+            .any(|(k, _v)| k.as_os_str() == OsStr::new("SCCACHE_RECACHE"));
         let cache_control = if force_recache {
             CacheControl::ForceRecache
         } else {
@@ -1141,6 +1174,7 @@ where
         let color_mode = hasher.color_mode();
         let me = self.clone();
         let kind = compiler.kind();
+        let lang = hasher.language();
         let creator = self.creator.clone();
         let storage = self.storage.clone();
         let pool = self.rt.clone();
@@ -1148,20 +1182,35 @@ where
         let task = async move {
             let dist_client = me.dist_client.get_client().await;
             let result = match dist_client {
-                Ok(client) => {
-                    hasher
-                        .get_cached_or_compile(
-                            client,
-                            creator,
-                            storage,
-                            arguments,
-                            cwd,
-                            env_vars,
-                            cache_control,
-                            pool,
+                Ok(client) => std::panic::AssertUnwindSafe(hasher.get_cached_or_compile(
+                    client,
+                    creator,
+                    storage,
+                    arguments,
+                    cwd,
+                    env_vars,
+                    cache_control,
+                    pool,
+                ))
+                .catch_unwind()
+                .await
+                .map_err(|e| {
+                    let panic = e
+                        .downcast_ref::<&str>()
+                        .map(|s| &**s)
+                        .or_else(|| e.downcast_ref::<String>().map(|s| &**s))
+                        .unwrap_or("An unknown panic was caught.");
+                    let thread = std::thread::current();
+                    let thread_name = thread.name().unwrap_or("unnamed");
+                    if let Some((file, line, column)) = PANIC_LOCATION.with(|l| l.take()) {
+                        anyhow!(
+                            "thread '{thread_name}' panicked at {file}:{line}:{column}: {panic}"
                         )
-                        .await
-                }
+                    } else {
+                        anyhow!("thread '{thread_name}' panicked: {panic}")
+                    }
+                })
+                .and_then(std::convert::identity),
                 Err(e) => Err(e),
             };
             let mut cache_write = None;
@@ -1171,16 +1220,22 @@ where
             };
             match result {
                 Ok((compiled, out)) => {
-                    let mut stats = me.stats.write().await;
+                    let mut stats = me.stats.lock().await;
                     match compiled {
                         CompileResult::Error => {
-                            stats.cache_errors.increment(&kind);
+                            debug!("compile result: cache error");
+
+                            stats.cache_errors.increment(&kind, &lang);
                         }
                         CompileResult::CacheHit(duration) => {
-                            stats.cache_hits.increment(&kind);
+                            debug!("compile result: cache hit");
+
+                            stats.cache_hits.increment(&kind, &lang);
                             stats.cache_read_hit_duration += duration;
                         }
                         CompileResult::CacheMiss(miss_type, dist_type, duration, future) => {
+                            debug!("compile result: cache miss");
+
                             match dist_type {
                                 DistType::NoDist => {}
                                 DistType::Ok(id) => {
@@ -1200,21 +1255,29 @@ where
                                     stats.cache_timeouts += 1;
                                 }
                                 MissType::CacheReadError => {
-                                    stats.cache_errors.increment(&kind);
+                                    stats.cache_errors.increment(&kind, &lang);
                                 }
                             }
-                            stats.cache_misses.increment(&kind);
-                            stats.cache_read_miss_duration += duration;
+                            stats.cache_misses.increment(&kind, &lang);
+                            stats.compiler_write_duration += duration;
+                            debug!("stats after compile result: {stats:?}");
                             cache_write = Some(future);
                         }
                         CompileResult::NotCacheable => {
-                            stats.cache_misses.increment(&kind);
+                            debug!("compile result: not cacheable");
+
+                            stats.cache_misses.increment(&kind, &lang);
                             stats.non_cacheable_compilations += 1;
                         }
                         CompileResult::CompileFailed => {
+                            debug!("compile result: compile failed");
+
                             stats.compile_fails += 1;
                         }
                     };
+                    // Make sure the write guard has been dropped ASAP.
+                    drop(stats);
+
                     let Output {
                         status,
                         stdout,
@@ -1229,7 +1292,7 @@ where
                     res.stderr = stderr;
                 }
                 Err(err) => {
-                    let mut stats = me.stats.write().await;
+                    let mut stats = me.stats.lock().await;
                     match err.downcast::<ProcessError>() {
                         Ok(ProcessError(output)) => {
                             debug!("Compilation failed: {:?}", output);
@@ -1261,7 +1324,7 @@ where
                                     error!("[{:?}] \t{}", out_pretty, e);
                                     let _ = writeln!(error, "sccache: caused by: {}", e);
                                 }
-                                stats.cache_errors.increment(&kind);
+                                stats.cache_errors.increment(&kind, &lang);
                                 //TODO: figure out a better way to communicate this?
                                 res.retcode = Some(-2);
                                 res.stderr = error.into_bytes();
@@ -1280,7 +1343,7 @@ where
                     match cache_write.await {
                         Err(e) => {
                             debug!("Error executing cache write: {}", e);
-                            me.stats.write().await.cache_write_errors += 1;
+                            me.stats.lock().await.cache_write_errors += 1;
                         }
                         //TODO: save cache stats!
                         Ok(info) => {
@@ -1289,7 +1352,7 @@ where
                                 info.object_file_pretty,
                                 util::fmt_duration_as_secs(&info.duration)
                             );
-                            let mut stats = me.stats.write().await;
+                            let mut stats = me.stats.lock().await;
                             stats.cache_writes += 1;
                             stats.cache_write_duration += info.duration;
                         }
@@ -1313,12 +1376,17 @@ where
 #[derive(Serialize, Deserialize, Debug, Clone, Default)]
 pub struct PerLanguageCount {
     counts: HashMap<String, u64>,
+    adv_counts: HashMap<String, u64>,
 }
 
 impl PerLanguageCount {
-    fn increment(&mut self, kind: &CompilerKind) {
-        let key = kind.lang_kind();
-        let count = self.counts.entry(key).or_insert(0);
+    fn increment(&mut self, kind: &CompilerKind, lang: &Language) {
+        let lang_comp_key = kind.lang_comp_kind(lang);
+        let adv_count = self.adv_counts.entry(lang_comp_key).or_insert(0);
+        *adv_count += 1;
+
+        let lang_key = kind.lang_kind(lang);
+        let count = self.counts.entry(lang_key).or_insert(0);
         *count += 1;
     }
 
@@ -1330,6 +1398,10 @@ impl PerLanguageCount {
         self.counts.get(key)
     }
 
+    pub fn get_adv(&self, key: &str) -> Option<&u64> {
+        self.adv_counts.get(key)
+    }
+
     pub fn new() -> PerLanguageCount {
         Self::default()
     }
@@ -1371,7 +1443,7 @@ pub struct ServerStats {
     /// The total time spent reading cache hits.
     pub cache_read_hit_duration: Duration,
     /// The total time spent reading cache misses.
-    pub cache_read_miss_duration: Duration,
+    pub compiler_write_duration: Duration,
     /// The count of compilation failures.
     pub compile_fails: u64,
     /// Counts of reasons why compiles were not cached.
@@ -1390,6 +1462,8 @@ pub struct ServerInfo {
     pub cache_location: String,
     pub cache_size: Option<u64>,
     pub max_cache_size: Option<u64>,
+    pub use_preprocessor_cache_mode: bool,
+    pub version: String,
 }
 
 /// Status of the dist client.
@@ -1421,7 +1495,7 @@ impl Default for ServerStats {
             cache_writes: u64::default(),
             cache_write_duration: Duration::new(0, 0),
             cache_read_hit_duration: Duration::new(0, 0),
-            cache_read_miss_duration: Duration::new(0, 0),
+            compiler_write_duration: Duration::new(0, 0),
             compile_fails: u64::default(),
             not_cached: HashMap::new(),
             dist_compiles: HashMap::new(),
@@ -1434,7 +1508,7 @@ impl ServerStats {
     /// Print stats to stdout in a human-readable format.
     ///
     /// Return the formatted width of each of the (name, value) columns.
-    fn print(&self) -> (usize, usize) {
+    fn print(&self, advanced: bool) -> (usize, usize) {
         macro_rules! set_stat {
             ($vec:ident, $var:expr, $name:expr) => {{
                 // name, value, suffix length
@@ -1452,6 +1526,16 @@ impl ServerStats {
                 }
             }};
         }
+        macro_rules! set_compiler_stat {
+            ($vec:ident, $var:expr, $name:expr) => {{
+                $vec.push(($name.to_string(), $var.all().to_string(), 0));
+                let mut sorted_stats: Vec<_> = $var.adv_counts.iter().collect();
+                sorted_stats.sort_by_key(|v| v.0);
+                for (lang, count) in sorted_stats.iter() {
+                    $vec.push((format!("{} ({})", $name, lang), count.to_string(), 0));
+                }
+            }};
+        }
 
         macro_rules! set_duration_stat {
             ($vec:ident, $dur:expr, $num:expr, $name:expr) => {{
@@ -1473,14 +1557,24 @@ impl ServerStats {
             self.requests_executed,
             "Compile requests executed"
         );
-        set_lang_stat!(stats_vec, self.cache_hits, "Cache hits");
-        set_lang_stat!(stats_vec, self.cache_misses, "Cache misses");
+        if advanced {
+            set_compiler_stat!(stats_vec, self.cache_hits, "Cache hits");
+            set_compiler_stat!(stats_vec, self.cache_misses, "Cache misses");
+        } else {
+            set_lang_stat!(stats_vec, self.cache_hits, "Cache hits");
+            set_lang_stat!(stats_vec, self.cache_misses, "Cache misses");
+        }
         set_stat!(stats_vec, self.cache_timeouts, "Cache timeouts");
         set_stat!(stats_vec, self.cache_read_errors, "Cache read errors");
         set_stat!(stats_vec, self.forced_recaches, "Forced recaches");
         set_stat!(stats_vec, self.cache_write_errors, "Cache write errors");
         set_stat!(stats_vec, self.compile_fails, "Compilation failures");
-        set_lang_stat!(stats_vec, self.cache_errors, "Cache errors");
+        if advanced {
+            set_compiler_stat!(stats_vec, self.cache_errors, "Cache errors");
+        } else {
+            set_lang_stat!(stats_vec, self.cache_errors, "Cache errors");
+        }
+
         set_stat!(
             stats_vec,
             self.non_cacheable_compilations,
@@ -1509,9 +1603,9 @@ impl ServerStats {
         );
         set_duration_stat!(
             stats_vec,
-            self.cache_read_miss_duration,
+            self.compiler_write_duration,
             self.cache_misses.all(),
-            "Average cache read miss"
+            "Average compiler"
         );
         set_duration_stat!(
             stats_vec,
@@ -1524,16 +1618,8 @@ impl ServerStats {
             self.dist_errors,
             "Failed distributed compilations"
         );
-        let name_width = stats_vec
-            .iter()
-            .map(|&(ref n, _, _)| n.len())
-            .max()
-            .unwrap();
-        let stat_width = stats_vec
-            .iter()
-            .map(|&(_, ref s, _)| s.len())
-            .max()
-            .unwrap();
+        let name_width = stats_vec.iter().map(|(n, _, _)| n.len()).max().unwrap();
+        let stat_width = stats_vec.iter().map(|(_, s, _)| s.len()).max().unwrap();
         for (name, stat, suffix_len) in stats_vec {
             println!(
                 "{:<name_width$} {:>stat_width$}",
@@ -1577,15 +1663,62 @@ impl ServerStats {
 }
 
 impl ServerInfo {
+    pub async fn new(stats: ServerStats, storage: Option<&dyn Storage>) -> Result<Self> {
+        let cache_location;
+        let use_preprocessor_cache_mode;
+        let cache_size;
+        let max_cache_size;
+        if let Some(storage) = storage {
+            cache_location = storage.location();
+            use_preprocessor_cache_mode = storage
+                .preprocessor_cache_mode_config()
+                .use_preprocessor_cache_mode;
+            (cache_size, max_cache_size) =
+                futures::try_join!(storage.current_size(), storage.max_size())?;
+        } else {
+            cache_location = String::new();
+            use_preprocessor_cache_mode = false;
+            cache_size = None;
+            max_cache_size = None;
+        }
+        let version = env!("CARGO_PKG_VERSION").to_string();
+        Ok(ServerInfo {
+            stats,
+            cache_location,
+            cache_size,
+            max_cache_size,
+            use_preprocessor_cache_mode,
+            version,
+        })
+    }
+
     /// Print info to stdout in a human-readable format.
-    pub fn print(&self) {
-        let (name_width, stat_width) = self.stats.print();
+    pub fn print(&self, advanced: bool) {
+        let (name_width, stat_width) = self.stats.print(advanced);
         println!(
             "{:<name_width$} {}",
             "Cache location",
             self.cache_location,
             name_width = name_width
         );
+        if self.cache_location.starts_with("Local disk") {
+            println!(
+                "{:<name_width$} {}",
+                "Use direct/preprocessor mode?",
+                if self.use_preprocessor_cache_mode {
+                    "yes"
+                } else {
+                    "no"
+                },
+                name_width = name_width
+            );
+        }
+        println!(
+            "{:<name_width$} {}",
+            "Version (client)",
+            self.version,
+            name_width = name_width
+        );
         for &(name, val) in &[
             ("Cache size", &self.cache_size),
             ("Max cache size", &self.max_cache_size),
@@ -1775,13 +1908,13 @@ impl Future for ShutdownOrInactive {
 /// Helper future which tracks the `ActiveInfo` below. This future will resolve
 /// once all instances of `ActiveInfo` have been dropped.
 struct WaitUntilZero {
-    info: std::sync::Weak<Mutex<Info>>,
+    info: std::sync::Weak<std::sync::Mutex<Info>>,
 }
 
 #[derive(Clone)]
 #[allow(dead_code)]
 struct ActiveInfo {
-    info: Arc<Mutex<Info>>,
+    info: Arc<std::sync::Mutex<Info>>,
 }
 
 struct Info {
@@ -1799,7 +1932,7 @@ impl Drop for Info {
 impl WaitUntilZero {
     #[rustfmt::skip]
     fn new() -> (WaitUntilZero, ActiveInfo) {
-        let info = Arc::new(Mutex::new(Info { waker: None }));
+        let info = Arc::new(std::sync::Mutex::new(Info { waker: None }));
 
         (WaitUntilZero { info: Arc::downgrade(&info) }, ActiveInfo { info })
     }
